﻿using System;
using Funq;
using ServiceStack.CacheAccess;
using ServiceStack.CacheAccess.Providers;
using ServiceStack.Common.Utils;
using ServiceStack.Configuration;
using ServiceStack.Logging;
using ServiceStack.Logging.Support.Logging;
using ServiceStack.WebHost.Endpoints;

using ICSV.Server;
using ICSV.Interface;

namespace ICSV.Host
{
    /// <summary>
    /// Initial configuration point of a ServiceStack web service application with a singleton AppHost.
    /// </summary>        
    public class AppHost : AppHostBase
    {
        private static ILog log;

        /// <summary>
        /// Initializes a new instance of the ServiceStack application, with the specified name and assembly containing the services.
        /// </summary>
        public AppHost()
            : base("ICSV", typeof(IntroService).Assembly)
        {
            LogManager.LogFactory = new ConsoleLogFactory();
            log = LogManager.GetLogger(typeof(AppHost));
        }

        /// <summary>
        /// Configures the ServiceStack application.
        /// </summary>
        public override void Configure(Container container)
        {
            //Permit modern browsers (e.g. Firefox) to allow sending of any REST HTTP Method
            base.SetConfig(new EndpointHostConfig
            {
                GlobalResponseHeaders =
                    {
                        { "Access-Control-Allow-Origin", "*" },
                        { "Access-Control-Allow-Methods", "GET, POST, PUT, DELETE, OPTIONS" },
                    },
            });

            container.Register<IResourceManager>(new ConfigurationResourceManager());
            container.Register(c => new EMDConfig(c.Resolve<IResourceManager>()));

            var appConfig = container.Resolve<EMDConfig>();

            appConfig.WebHostImgPath = System.Web.HttpContext.Current.Server.MapPath("~/img/");

            //The MemoryCacheClient is a great way to get started with caching; nothing external to setup.
            container.Register<ICacheClient>(c => new MemoryCacheClient());

            log.InfoFormat("AppHost Configured: {0}", DateTime.Now);
        }
    }

}