﻿/*
Copyright (c) 2015 Stefan Spelitz
This software is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any damages
arising from the use of this software.
Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:
1. The origin of this software must not be misrepresented; you must not
claim that you wrote the original software. If you use this software
in a product, an acknowledgment in the product documentation would be
appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must not
be misrepresented as being the original software.
3. This notice may not be removed or altered from any source
distribution.
*/
/// <reference path="js/three.d.ts"/>
/// <reference path="js/jquery.d.ts"/>
var serverURL = location.protocol + "//" + location.host + '/servicestack/';

//var serverURL = 'http://testerosa.apphb.com/servicestack';
var wrapper;

var SmartPhone;

var ThreeWrapper = (function () {
    function ThreeWrapper(element) {
        this.INITIAL_CAM_POS_FRONT = new THREE.Vector3(0, -100, 500);
        this.AMBIENT_OCCLUSION_BORDER_IN_PIXEL = 20;
        this.AMBIENT_OCCLUSION_INTENSITY = 1;
        this.IMAGE_SIZE = 40;
        this.MINIMUM_CAMERA_DISTANCE = 180;
        this.HUD_CUBE_SIZE = 500;
        this.ZOOM_SPEED_ON_MOBILE = 0.35;
        this.MAX_SERVER_ERRORS_ALLOWED = 10;
        this.lastContainerWidth = 0;
        this.lastContainerHeight = 0;
        this.cubeSides = [];
        this.cubeSideSelected = undefined;
        this.cubeSideSelectedOnMouseDown = undefined;
        this.objectsMap = {};
        this.objectScaleMap = {};
        this.objectPositionsMap3D = {};
        this.objectPositionsMap2D = {};
        this.objectPositionsMap3DPrevious = {};
        // Ambient occlusion
        this.objectsBackMap = {};
        this.objectsBackScaleMap = {};
        this.currentAxesList = [];
        this.axes3D = [];
        this.axes2D = [];
        this.searching = false;
        this.cancelled = false;
        this.searchingForCount = 0;
        this.currentSearchId = -1;
        this.serverErrorsSoFar = 0;
        this.view3D = true;
        this.showAnimations = true;
        this.showShadows = true;
        this.showShadowsForThisSearch = true;
        // spacing:
        this.spacingFactor = 1;
        this.spacingFactorUser = 1;
        this.spacingFactorUser2D = 1;
        this.spacingFactorUser3D = 1;
        this.lastCameraObjectDistance = 50000;
        this.tweenMap = {};
        // raycasting:
        this.mousePosition = new THREE.Vector2();
        this.raycaster = new THREE.Raycaster();
        this.zoomInObjectId = undefined;
        this.zoomInObjectIdPrevious = undefined;
        this.selectedSpriteId = undefined;
        this.selectedSpriteIdPrev = undefined;
        this.mouseMoveOccurred = false;
        this.gettingOutOfZoom = false;
        // mobile platform - zooming:
        this.dollyStart = new THREE.Vector2();
        this.dollyEnd = new THREE.Vector2();
        this.dollyDelta = new THREE.Vector2();
        this.element = element;
    }
    ThreeWrapper.prototype.init = function () {
        SmartPhone.setUserAgent(navigator.userAgent);

        // CONFIGURE WEBGL / THREE.js:
        // Also see: http://webglfundamentals.org/webgl/lessons/webgl-anti-patterns.html
        var canvas = document.getElementById('canvas');
        this.canvasContainer = document.getElementById('canvasContainer');
        this.topControls = document.getElementById('topControls');
        this.jQueryContainer = $("#canvasContainer");
        this.jQueryCanvas = $("canvas");
        this.canvas = canvas;

        this.renderer = new THREE.WebGLRenderer({ canvas: canvas, antialias: true });
        this.renderer.setPixelRatio(window.devicePixelRatio);
        this.renderer.setViewport(0, 0, this.renderer.context.drawingBufferWidth, this.renderer.context.drawingBufferHeight);

        this.renderer.setClearColor(0xffffff, 1);
        this.renderer.autoClear = false;

        // anisotropic filtering:
        var maxAnisotropy = this.renderer.getMaxAnisotropy();

        var aspect = canvas.clientWidth / canvas.clientHeight;
        var nearPlane = 20;

        var camerapos = this.INITIAL_CAM_POS_FRONT;
        this.camera3D = new THREE.PerspectiveCamera(75, aspect, nearPlane, 5000);
        this.camera3D.position.copy(camerapos);
        this.camera3D.lookAt(new THREE.Vector3(0, 0, 0));

        this.camera = this.camera3D;

        var orthoSize = 2200;
        this.camera2D = new THREE.OrthographicCamera(-orthoSize / 2, orthoSize / 2, orthoSize / 2 / aspect, -orthoSize / 2 / aspect, nearPlane, 5000);
        this.camera2D.position.set(camerapos.x, 0, camerapos.z);
        this.camera2D.lookAt(new THREE.Vector3(0, 0, 0));

        this.cameraHud = new THREE.PerspectiveCamera(75, aspect, 0.1, 5000);
        this.cameraHud.position.copy(camerapos);
        this.cameraHud.lookAt(new THREE.Vector3());
        this.cameraHud.setViewOffset(canvas.clientWidth / 4, canvas.clientHeight / 4, 0, 0, canvas.clientWidth, canvas.clientHeight);

        this.scene = new THREE.Scene();
        this.spriteContainer = new THREE.Object3D();
        this.scene.add(this.spriteContainer);

        //this.scene.fog = new THREE.FogExp2(0xaaaaaa, 0.0025);
        this.sceneHud = new THREE.Scene();

        // geometry for sprites (to reuse)
        this.spriteGeometry = new THREE.PlaneBufferGeometry(1, 1, 2, 2);

        // red selection plane:
        var pmaterial = new THREE.MeshBasicMaterial({ color: 0xff0000, side: THREE.DoubleSide });
        var pgeometry = new THREE.PlaneBufferGeometry(1, 1, 2, 2);
        this.selectionBorder = new THREE.Mesh(pgeometry, pmaterial);
        this.selectionBorder.visible = false;
        this.scene.add(this.selectionBorder);

        this.spriteBackContainer = new THREE.Object3D();
        this.scene.add(this.spriteBackContainer);

        this.axesContainer = new THREE.Object3D();
        this.scene.add(this.axesContainer);

        // setup server connection:
        this.gateway = new servicestack.ClientGateway(serverURL);

        // cube
        var tex1 = THREE.ImageUtils.loadTexture('img/cube/right.png');
        var tex2 = THREE.ImageUtils.loadTexture('img/cube/left.png');
        var tex3 = THREE.ImageUtils.loadTexture('img/cube/top.png');
        var tex4 = THREE.ImageUtils.loadTexture('img/cube/bottom.png');
        var tex5 = THREE.ImageUtils.loadTexture('img/cube/front.png');
        var tex6 = THREE.ImageUtils.loadTexture('img/cube/back.png');
        var tex7 = THREE.ImageUtils.loadTexture('img/cube/2D.png');
        tex1.anisotropy = maxAnisotropy;
        tex2.anisotropy = maxAnisotropy;
        tex3.anisotropy = maxAnisotropy;
        tex4.anisotropy = maxAnisotropy;
        tex5.anisotropy = maxAnisotropy;
        tex6.anisotropy = maxAnisotropy;
        tex7.anisotropy = maxAnisotropy;
        tex7.minFilter = tex7.magFilter = THREE.LinearFilter;

        var faces2D = [
            new THREE.MeshBasicMaterial({ map: tex1 }),
            new THREE.MeshBasicMaterial({ map: tex2 }),
            new THREE.MeshBasicMaterial({ map: tex3 }),
            new THREE.MeshBasicMaterial({ map: tex4 }),
            new THREE.MeshBasicMaterial({ map: tex7 }),
            new THREE.MeshBasicMaterial({ map: tex6 })
        ];

        // old cube
        var geometry = new THREE.BoxGeometry(300, 300, 300);

        //var faces3D = [
        //    new THREE.MeshBasicMaterial({ map: tex1 }),
        //    new THREE.MeshBasicMaterial({ map: tex2 }),
        //    new THREE.MeshBasicMaterial({ map: tex3 }),
        //    new THREE.MeshBasicMaterial({ map: tex4 }),
        //    new THREE.MeshBasicMaterial({ map: tex5 }),
        //    new THREE.MeshBasicMaterial({ map: tex6 })
        //];
        //var material = new THREE.MeshFaceMaterial(faces3D);
        //this.cube = new THREE.Mesh(geometry, material);
        //this.sceneHud.add(this.cube);
        // cube with planes (for raycasting)
        var cubePlaneGeom = new THREE.PlaneBufferGeometry(300, 300, 2, 2);

        var rightMaterial = new THREE.MeshBasicMaterial({ map: tex1 });
        var leftMaterial = new THREE.MeshBasicMaterial({ map: tex2 });
        var topMaterial = new THREE.MeshBasicMaterial({ map: tex3 });
        var bottomMaterial = new THREE.MeshBasicMaterial({ map: tex4 });
        var frontMaterial = new THREE.MeshBasicMaterial({ map: tex5 });
        var backMaterial = new THREE.MeshBasicMaterial({ map: tex6 });

        var rightPlane = new THREE.Mesh(cubePlaneGeom, rightMaterial);
        var leftPlane = new THREE.Mesh(cubePlaneGeom, leftMaterial);
        var topPlane = new THREE.Mesh(cubePlaneGeom, topMaterial);
        var bottomPlane = new THREE.Mesh(cubePlaneGeom, bottomMaterial);
        var frontPlane = new THREE.Mesh(cubePlaneGeom, frontMaterial);
        var backPlane = new THREE.Mesh(cubePlaneGeom, backMaterial);

        this.cubeSides = [rightPlane, leftPlane, topPlane, bottomPlane, frontPlane, backPlane];

        rightPlane.rotation.y = Math.PI / 2;
        leftPlane.rotation.y = -Math.PI / 2;
        backPlane.rotation.y = Math.PI;
        topPlane.rotation.x = -Math.PI / 2;
        bottomPlane.rotation.x = Math.PI / 2;

        rightPlane.position.x = 150;
        leftPlane.position.x = -150;
        topPlane.position.y = 150;
        bottomPlane.position.y = -150;
        frontPlane.position.z = 150;
        backPlane.position.z = -150;

        var virtualCube = new THREE.Object3D();
        virtualCube.add(rightPlane);
        virtualCube.add(leftPlane);
        virtualCube.add(topPlane);
        virtualCube.add(bottomPlane);
        virtualCube.add(frontPlane);
        virtualCube.add(backPlane);

        this.sceneHud.add(virtualCube);
        this.cube = virtualCube;

        var material = new THREE.MeshFaceMaterial(faces2D);
        this.cube2D = new THREE.Mesh(geometry, material);
        this.sceneHud.add(this.cube2D);
        this.cube2D.visible = false;

        // plane
        var gridTex1 = THREE.ImageUtils.loadTexture('img/grid2.png');
        gridTex1.wrapS = gridTex1.wrapT = THREE.RepeatWrapping;
        gridTex1.repeat.set(30, 30);
        gridTex1.minFilter = THREE.LinearMipMapLinearFilter;
        gridTex1.magFilter = THREE.LinearMipMapLinearFilter;
        gridTex1.anisotropy = maxAnisotropy;

        var pmaterial = new THREE.MeshBasicMaterial({ map: gridTex1, transparent: true, opacity: 1, side: THREE.DoubleSide });
        var pgeometry = new THREE.PlaneBufferGeometry(5000, 5000, 100, 100);
        var planeMesh = new THREE.Mesh(pgeometry, pmaterial);
        planeMesh.rotation.x = -Math.PI / 2;
        planeMesh.position.y = -300;
        this.scene.add(planeMesh);
        this.groundPlane = planeMesh;

        // background grid (for 2D)
        var gridTex = THREE.ImageUtils.loadTexture('img/grid.png');
        gridTex.wrapS = gridTex.wrapT = THREE.RepeatWrapping;
        gridTex.repeat.set(60, 60);
        gridTex.minFilter = THREE.LinearMipMapLinearFilter;
        gridTex.magFilter = THREE.LinearMipMapLinearFilter;

        var gridMaterial = new THREE.MeshBasicMaterial({ map: gridTex, transparent: true, opacity: 1 });
        var gridGeometry = new THREE.PlaneBufferGeometry(5000, 5000, 30, 30);
        var grid = new THREE.Mesh(gridGeometry, gridMaterial);
        grid.position.z = -500;
        grid.visible = false;
        this.scene.add(grid);
        this.backgroundGrid = grid;

        //
        this.controls3D = new THREE.MyOrbitControls(this.camera3D, this.renderer.domElement);

        //this.controls.maxPolarAngle = 3 * Math.PI / 4;
        this.controls3D.noZoom = true;

        //this.controls.minDistance = 0;
        //this.controls.maxDistance = Infinity;
        //this.controls.rotateSpeed = 0.5;
        if (SmartPhone.isAny()) {
            // slower rotation with touch:
            this.controls3D.rotateSpeed = 0.7;
        }

        this.controls = this.controls3D;

        this.controls2D = new THREE.MyOrbitControls(this.camera2D, this.renderer.domElement);

        //this.controls.maxPolarAngle = 3 * Math.PI / 4;
        this.controls2D.noZoom = true;
        this.controls2D.noRotate = true;
        this.controls2D.enabled = false;

        //this.controls.minDistance = 0;
        //this.controls.maxDistance = Infinity;
        //this.controls.rotateSpeed = 0.5;
        //
        this.shadowTexture = THREE.ImageUtils.loadTexture('img/shadow.png');
        this.shadowTexture.minFilter = THREE.LinearFilter;
        this.shadowTexture.magFilter = THREE.LinearFilter;

        this.selectionTexture = THREE.ImageUtils.loadTexture('img/selection.png');
        this.selectionTexture.minFilter = THREE.LinearFilter;
        this.selectionTexture.magFilter = THREE.LinearFilter;

        //
        window.onbeforeunload = function unloadFunction() {
            wrapper.onWindowClose();
            if (wrapper.searching)
                return '';
            return null;
        };

        //this.renderer.domElement.addEventListener('dblclick', this.onMouseDblClick, false);
        this.canvasContainer.addEventListener('mousemove', this.onMouseMove, false);
        this.renderer.domElement.addEventListener('mousedown', this.onMouseDown, false);
        this.renderer.domElement.addEventListener('mouseup', this.onMouseUp, false);

        this.renderer.domElement.addEventListener('mousewheel', this.onMouseWheel, false);
        this.renderer.domElement.addEventListener('DOMMouseScroll', this.onMouseWheel, false); // firefox

        this.renderer.domElement.addEventListener('touchstart', this.onTouchStart, false);
        this.renderer.domElement.addEventListener('touchmove', this.onTouchMove, false);

        //this.renderer.domElement.addEventListener('touchend', this.onAnythingToUndoZoomIn, false);
        window.addEventListener('keydown', this.onAnythingToUndoZoomIn, false);

        this.configureGUI();

        this.reset();
    };

    ThreeWrapper.prototype.configureGUI = function () {
        // ----------- FOR MOBILE -----------
        if (SmartPhone.isAny()) {
            // disable shadows and animations per default
            document.getElementById('checkAnim').checked = false;
            document.getElementById('checkShadow').checked = false;
            this.showAnimations = false;
            this.showShadows = false;
        }

        // ----------- SLIDERS -----------
        var $document = $(document);
        var $r = $('input[type=range]');
        var sliderImgCount = document.getElementById('numImgCount');
        var labImgCount = document.getElementById('labImgCount');
        var sliderSpacing = document.getElementById('numSpacing');
        var labSpacing = document.getElementById('labSpacing');

        // set initial output value
        labImgCount.innerHTML = sliderImgCount.value;

        // update output value
        $document.on('input', 'input[type="range"]', function (e) {
            labImgCount.innerHTML = sliderImgCount.value;
            wrapper.updateForSpacingSlider(sliderSpacing.value);
        });

        // Initialize
        $r.rangeslider({
            polyfill: false
        });
    };

    ThreeWrapper.prototype.createAxis = function (x1, y1, z1, color1, x2, y2, z2, color2) {
        var cylinderColor = 0x888888;
        var cylinderThickness = 1.5;
        var headLength = 25;
        var headWidth = 10;

        var cmaterial = new THREE.MeshBasicMaterial({ color: cylinderColor });
        var cylinderMesh = new THREE.Mesh(new THREE.CylinderGeometry(cylinderThickness, cylinderThickness, 1, 50, 50, false), cmaterial);
        var cylinder = new THREE.Object3D();
        cylinder.add(cylinderMesh);

        // arrows:
        var p1 = new THREE.Vector3(x1, y1, z1);
        var p2 = new THREE.Vector3(x2, y2, z2);
        var direction = new THREE.Vector3().subVectors(p2, p1);
        var directionNorm = direction.clone().normalize();

        var arrow1 = new THREE.ArrowHelper(directionNorm, p1, direction.length(), color2, headLength, headWidth);
        var arrow2 = new THREE.ArrowHelper(directionNorm.clone().multiplyScalar(-1), p2, direction.length(), color1, headLength, headWidth);

        var cylinderPos = new THREE.Vector3().addVectors(p1, direction.clone().multiplyScalar(0.5));
        cylinder.position.set(cylinderPos.x, cylinderPos.y, cylinderPos.z);
        cylinder.rotation.set(arrow1.rotation.x, arrow1.rotation.y, arrow1.rotation.z);
        cylinder.scale.y = direction.length() - headLength * 2;

        var axis = new THREE.Object3D();
        axis.add(arrow1);
        axis.add(arrow2);
        axis.add(cylinder);

        return axis;
    };

    ThreeWrapper.prototype.reset = function () {
        for (var key in this.objectsMap) {
            if (this.objectsMap.hasOwnProperty(key)) {
                this.spriteContainer.remove(this.objectsMap[key]);
                this.spriteBackContainer.remove(this.objectsBackMap[key]);
            }
        }

        // clear axes:
        this.clearCurrentAxes();
        this.axes2D = [];
        this.axes3D = [];

        this.objectsMap = {};
        this.objectScaleMap = {};
        this.objectsBackMap = {};
        this.objectPositionsMap3D = {};
        this.objectPositionsMap2D = {};
        this.tweenMap = {};
        this.zoomInObjectId = undefined;
        this.zoomInObjectIdPrevious = undefined;
        this.zoomInObjectPosition = undefined;
        this.spacingFactor = 1;
        this.spacingFactorUser = 1;
        this.spacingFactorUser2D = 1;
        this.spacingFactorUser3D = 1;
        this.lastCameraObjectDistance = 50000;
        this.selectedSpriteId = undefined;
        this.selectedSpriteIdPrev = undefined;
        this.searching = false;
        this.searchingForCount = 0;
        this.currentSearchId = -1;
        this.cancelled = false;
        this.serverErrorsSoFar = 0;

        this.showAnimations = document.getElementById('checkAnim').checked;
        this.showShadows = document.getElementById('checkShadow').checked;

        // restore states:
        $("#btnSearch").show();
        $("#btnCancel").hide();
        $("#imgSearch").css('visibility', 'hidden');
        document.getElementById("txtSearch").disabled = false;
        document.getElementById('numSpacing').value = 15;
        $('input[type=range]').rangeslider('update', true);

        // add status text:
        this.log('Status: Start a new search query.' + '<br/>' + '-------------------------');

        this.displayImageCount();
        this.updateSpacingGUI();
    };

    ThreeWrapper.prototype.displayImageCount = function () {
        var div = document.getElementById('txtImageLoaded');
        var objectCount = Object.keys(this.objectsMap).length;

        if (this.cancelled) {
            div.innerHTML = "Cancelled: [" + objectCount + " / " + this.searchingForCount + "]";
        } else if (!this.searching && objectCount == 0 && this.searchingForCount > 0) {
            div.innerHTML = "No results";
        } else if (!this.searching && objectCount > 0 && this.searchingForCount > 0) {
            div.innerHTML = "Finished: [" + objectCount + " / " + this.searchingForCount + "]";
        } else if (this.searching) {
            div.innerHTML = "Searching... [" + objectCount + " / " + this.searchingForCount + "]";
        } else {
            div.innerHTML = "Ready...";
        }
    };

    ThreeWrapper.prototype.switch2d3d = function () {
        this.undoZoomIn();

        this.view3D = !this.view3D;

        // change axes
        this.updateAxes();

        // switch cameras
        if (this.view3D) {
            this.controls = this.controls3D;
            this.camera = this.camera3D;
            this.controls2D.enabled = false;
            this.controls3D.enabled = true;
        } else {
            this.controls = this.controls2D;
            this.camera = this.camera2D;
            this.controls2D.enabled = true;
            this.controls3D.enabled = false;
        }

        // switch HUD
        this.cube.visible = this.view3D;
        this.cube2D.visible = !this.view3D;

        // change visibility of background grid
        this.backgroundGrid.visible = !this.view3D;

        // switch spacing:
        if (this.view3D) {
            this.spacingFactorUser2D = this.spacingFactorUser;
            this.spacingFactorUser = this.spacingFactorUser3D;
        } else {
            this.spacingFactorUser3D = this.spacingFactorUser;
            this.spacingFactorUser = this.spacingFactorUser2D;
        }

        if (this.view3D) {
            // restore 3D positions:
            this.changeSpacing(this.spacingFactor);
            for (var key in this.objectsMap) {
                if (this.objectsMap.hasOwnProperty(key)) {
                    var sprite = this.objectsMap[key];
                    if (this.objectPositionsMap3DPrevious.hasOwnProperty(key)) {
                        var pos = this.objectPositionsMap3DPrevious[key];
                        this.transitionForSpacing(sprite, key, pos, true);
                    } else {
                        var posCalc = this.calculateCurrentPositionWithCamera(key);
                        this.transitionForSpacing(sprite, key, posCalc, true);
                    }
                }
            }
        } else {
            // remember 3D positions:
            this.objectPositionsMap3DPrevious = {};
            for (var key in this.objectsMap) {
                if (this.objectsMap.hasOwnProperty(key)) {
                    var sprite = this.objectsMap[key];
                    var pos = sprite.position.clone();
                    if (key == this.zoomInObjectId) {
                        // the zoom in object needs special handling -> don't save the current position, but the 'should-be-position'
                        pos = this.zoomInObjectPosition;
                    }
                    this.objectPositionsMap3DPrevious[key] = pos;
                }
            }
            this.separateAccordingToCamera();
        }

        // update button icon
        var img3D = $('#img3D');
        img3D.prop('src', this.view3D ? 'img/2D_red.png' : 'img/3D_green.png');

        this.updateSpacingGUI();
    };

    ThreeWrapper.prototype.increaseSpacing = function () {
        // update slider:
        var sliderSpacing = document.getElementById('numSpacing');
        var value = sliderSpacing.value;
        if (value == sliderSpacing.max)
            return;
        sliderSpacing.value++;

        $('input[type=range]').rangeslider('update', true);

        this.spacingFactorUser *= 8; // 2 for sqrt, 8 for log
        this.changeSpacing(this.spacingFactor);
    };

    ThreeWrapper.prototype.decreaseSpacing = function () {
        // update slider:
        var sliderSpacing = document.getElementById('numSpacing');
        var value = sliderSpacing.value;
        if (value == sliderSpacing.min)
            return;
        sliderSpacing.value--;

        $('input[type=range]').rangeslider('update', true);

        this.spacingFactorUser /= 8; // 2 for sqrt, 8 for log
        this.changeSpacing(this.spacingFactor);
    };

    ThreeWrapper.prototype.updateForSpacingSlider = function (sliderSpacing) {
        // min: 0, max: 31, 100 % = 15
        sliderSpacing -= 15;
        this.spacingFactorUser = Math.pow(8, sliderSpacing);
        this.changeSpacing(this.spacingFactor);
    };

    ThreeWrapper.prototype.changeSpacing = function (newFactor) {
        this.spacingFactor = newFactor;
        this.updateSpacingGUI();

        for (var key in this.objectsMap) {
            if (this.objectsMap.hasOwnProperty(key)) {
                var sprite = this.objectsMap[key];
                var posNew = this.calculateCurrentPositionWithCamera(key);
                this.transitionForSpacing(sprite, key, posNew, true);
            }
        }
    };

    ThreeWrapper.prototype.updateSpacingGUI = function () {
        var factor = this.spacingFactor * this.spacingFactorUser;
        if (factor < 1) {
            factor = 1 / (1 + Math.log(1 / factor));
        } else {
            factor = 1 + Math.log(factor);
        }

        var label = document.getElementById('labSpacing');
        label.innerHTML = Math.round(factor * 100) + ' %';

        var slider = document.getElementById('numSpacing');
        slider.value = Math.round(Math.log(this.spacingFactorUser) / Math.log(8)) + 15;
        $('input[type=range]').rangeslider('update', true);
    };

    ThreeWrapper.prototype.cancelImageSearch = function () {
        if (!this.searching)
            return;
        var searchId = this.currentSearchId;
        this.gateway.getFromService({
            CancelSearch: { ForSearchId: searchId }
        }, function (e) {
            // nothing to do
        }, function (status) {
            wrapper.logServerError(status, undefined);
        });
    };

    ThreeWrapper.prototype.triggerImageSearch = function (queryString, photoCount) {
        if (this.searching)
            return;
        this.reset();
        this.searching = true;
        this.showShadowsForThisSearch = this.showShadows;

        $("#btnSearch").hide();
        $("#btnCancel").show();
        $("#imgSearch").css('visibility', 'visible');
        document.getElementById("txtSearch").disabled = true;
        $("#shadowHint").css('visibility', 'hidden');

        this.searchingForCount = photoCount;
        this.displayImageCount();

        this.gateway.getFromService({
            SearchImages: { ForQuery: queryString, PhotoCount: photoCount }
        }, function (e) {
            wrapper.currentSearchId = e.SearchId;
            wrapper.pollStatus(e.SearchId);
            wrapper.pollResults(e.SearchId);
        }, function (status) {
            wrapper.logServerError(status, undefined);
            alert('Error connecting to server. See log for details.');
            wrapper.imageSearchFinished();
        });
    };

    ThreeWrapper.prototype.logServerError = function (status, attempt) {
        if (status.ErrorCode == undefined && status.ErrorMessage == undefined && status.Message == undefined) {
            if (attempt == undefined) {
                this.log('Unknown server error. Either connection failed or timeout.');
            } else {
                wrapper.log('(Error #' + attempt + ') Unknown server error. Either connection failed or timeout.');
            }
        } else {
            if (attempt == undefined) {
                wrapper.log('Server error: [' + status.ErrorCode + '] Message: ' + status.ErrorMessage + ' (' + status.Message + ')');
            } else {
                wrapper.log('(Error # ' + attempt + ') Server error: [' + status.ErrorCode + '] Message: ' + status.ErrorMessage + ' (' + status.Message + ')');
            }
        }
    };

    ThreeWrapper.prototype.imageSearchFinished = function () {
        if (!this.searching)
            return;
        this.searching = false;
        $("#btnSearch").show();
        $("#btnCancel").hide();
        $("#imgSearch").css('visibility', 'hidden');
        document.getElementById("txtSearch").disabled = false;
        this.displayImageCount();
    };

    ThreeWrapper.prototype.pollStatus = function (searchId) {
        this.gateway.getFromService({
            GetStatus: { ForSearchId: searchId }
        }, function (e) {
            wrapper.processStatus(searchId, e.Status);
        }, function (status) {
            wrapper.serverErrorsSoFar++;
            var attempt = wrapper.serverErrorsSoFar;
            wrapper.logServerError(status, attempt);

            if (wrapper.serverErrorsSoFar >= wrapper.MAX_SERVER_ERRORS_ALLOWED) {
                alert('One or more errors occurred while trying to read from server. See log for details.');
                wrapper.cancelImageSearch(); // send cancel signal
                wrapper.imageSearchFinished();
            } else {
                // try again
                wrapper.pollStatus(searchId);
            }
        });
    };

    ThreeWrapper.prototype.pollResults = function (searchId) {
        this.gateway.getFromService({
            GetImageResults: { ForSearchId: searchId }
        }, function (e) {
            wrapper.axes3D = e.Axes3D;
            wrapper.axes2D = e.Axes2D;
            wrapper.updateAxes(); // recreate all axes

            // handle images:
            if (e.IsFinalResult) {
                wrapper.displayImages(e.Images);
                wrapper.imageSearchFinished();
                // nothing more to do
            } else {
                wrapper.displayImages(e.Images);
                wrapper.pollResults(e.SearchId);
                // poll again! more results availabe at server side
            }
        }, function (status) {
            wrapper.serverErrorsSoFar++;
            var attempt = wrapper.serverErrorsSoFar;
            wrapper.logServerError(status, attempt);

            if (wrapper.serverErrorsSoFar >= wrapper.MAX_SERVER_ERRORS_ALLOWED) {
                alert('One or more errors occurred while trying to read from server. See log for details.');
                wrapper.cancelImageSearch(); // send cancel signal
                wrapper.imageSearchFinished();
            } else {
                // try again
                wrapper.pollResults(searchId);
            }
        });
    };

    ThreeWrapper.prototype.processStatus = function (searchId, status) {
        var finished = false;
        var cancelled = false;
        for (var i = 0; i < status.length; i++) {
            this.log(status[i].Message);
            if (status[i].Id == 2) {
                finished = true;
            } else if (status[i].Id == 3) {
                cancelled = true;
                wrapper.cancelled = true;
                finished = true;
            }
        }
        if (!finished) {
            wrapper.pollStatus(searchId);
        }
        if (cancelled) {
            wrapper.displayImageCount();
        }
    };

    ThreeWrapper.prototype.log = function (message) {
        var div = document.getElementById('statustext');
        div.innerHTML = message + '<br/>' + div.innerHTML;
    };

    ThreeWrapper.prototype.clearCurrentAxes = function () {
        for (var i = 0; i < this.currentAxesList.length; i++) {
            this.axesContainer.remove(this.currentAxesList[i]);
        }
        this.currentAxesList = [];
    };

    ThreeWrapper.prototype.updateAxes = function () {
        var axes;
        if (this.view3D) {
            axes = this.axes3D;
        } else {
            axes = this.axes2D;
        }
        if (axes.length == 0)
            return;

        // HACK: put axes in 2D mode a bit to the back to avoid overlapping with images
        var zOffset = 0;
        if (!this.view3D) {
            zOffset = -3;
        }

        this.clearCurrentAxes();

        for (var i = 0; i < axes.length; i++) {
            if (!this.view3D && i == 1)
                continue;

            var axisData = axes[i];
            var axis = this.createAxis(axisData.XCoordP1, axisData.YCoordP1, axisData.ZCoordP1 + zOffset, axisData.ColorP1, axisData.XCoordP2, axisData.YCoordP2, axisData.ZCoordP2 + zOffset, axisData.ColorP2);
            this.axesContainer.add(axis);
            this.currentAxesList.push(axis);
        }
    };

    ThreeWrapper.prototype.displayImages = function (images) {
        for (var i = 0; i < images.length; i++) {
            var objectId = images[i].Id;
            if (objectId == null || objectId == undefined)
                continue;

            var imagePos = images[i].pos3D;
            var finalPosition = new THREE.Vector3(imagePos.XCoord, imagePos.YCoord, imagePos.ZCoord);
            this.objectPositionsMap3D[objectId] = finalPosition;

            var imagePos = images[i].pos2D;
            var finalPosition = new THREE.Vector3(imagePos.XCoord, imagePos.YCoord, imagePos.ZCoord);
            this.objectPositionsMap2D[objectId] = finalPosition;

            if (objectId in this.objectsMap) {
                // update for existing data
                // -> make transition to new location
                var object = this.objectsMap[objectId];
                this.transitionForSpacing(object, objectId, this.calculateCurrentPositionWithCamera(objectId), true);
            } else {
                // new data
                // put dummy objects in map, until the sprite is fully loaded:
                this.objectsMap[objectId] = new THREE.Object3D();
                this.objectsBackMap[objectId] = new THREE.Object3D();
                this.objectsBackScaleMap[objectId] = new THREE.Vector3();
                this.loadImage(images[i].Url, objectId, this.createSprite);
            }
        }
        this.displayImageCount();
    };

    ThreeWrapper.prototype.createSpriteInternal = function (imageId, imageElement) {
        var tex = new THREE.Texture(imageElement);
        tex.needsUpdate = true;
        tex.minFilter = THREE.LinearFilter;
        tex.magFilter = THREE.LinearFilter;

        var factorX, factorY;
        var borderInPixel = this.AMBIENT_OCCLUSION_BORDER_IN_PIXEL;
        factorY = borderInPixel / imageElement.height;
        factorX = borderInPixel / imageElement.width;

        // the sprites don't even need a texture, we just need them for the positions where to render (for the sprites with shadows)
        var pmaterial = new THREE.MeshBasicMaterial({ color: 0xffffff });
        var pgeometry = this.spriteGeometry;
        var sprite = new THREE.Mesh(pgeometry, pmaterial);

        // the actual sprites (i.e. texture + shadow surrounding the texture) are created here:
        var pbackmaterial;
        if (this.showShadowsForThisSearch) {
            pbackmaterial = new THREE.ShaderMaterial({
                uniforms: {
                    spriteTex: { type: 't', value: tex },
                    shadowTex: { type: 't', value: this.shadowTexture },
                    factorX: { type: 'f', value: factorX },
                    factorY: { type: 'f', value: factorY },
                    darken: { type: 'f', value: this.AMBIENT_OCCLUSION_INTENSITY }
                },
                vertexShader: document.getElementById('vertShader').textContent,
                fragmentShader: document.getElementById('fragShader').textContent,
                transparent: true
            });
        } else {
            pbackmaterial = new THREE.MeshBasicMaterial({ map: tex });
        }
        var pbackgeometry = this.spriteGeometry;
        var spriteBack = new THREE.Mesh(pbackgeometry, pbackmaterial);

        // random start position:
        sprite.position.x = Math.random() * 4000 - 2000, sprite.position.y = Math.random() * 4000 - 2000, sprite.position.z = Math.random() * 4000 - 2000;

        var spriteSize = this.IMAGE_SIZE;
        var longEdge = Math.max(imageElement.width, imageElement.height);
        var scaleFactor = longEdge / spriteSize;

        sprite.scale.set(imageElement.width / scaleFactor, imageElement.height / scaleFactor, 1);

        // the sprites (with shadow) are scaled, s.t. the inner part (i.e. the sprite texture) is exactly as big as the sprite
        // and the shadow part is the excess space surrounding the sprite texture
        // YEAH - a bit of math:
        var scaleX = imageElement.width / (imageElement.width - borderInPixel * 2);
        var scaleY = imageElement.height / (imageElement.height - borderInPixel * 2);
        spriteBack.scale.set(sprite.scale.x * scaleX, sprite.scale.y * scaleY, 1);

        this.objectScaleMap[imageId] = sprite.scale.clone();
        this.objectsMap[imageId] = sprite;
        this.spriteContainer.add(sprite);

        this.objectsBackMap[imageId] = spriteBack;
        this.objectsBackScaleMap[imageId] = new THREE.Vector3(scaleX, scaleY, 1);
        this.spriteBackContainer.add(spriteBack);

        var finalPosition = this.calculateCurrentPositionWithCamera(imageId);
        this.transitionRandom(sprite, imageId, finalPosition);
    };

    ThreeWrapper.prototype.createSprite = function (imageId, imageElement) {
        wrapper.createSpriteInternal(imageId, imageElement);
    };

    ThreeWrapper.prototype.loadImage = function (imageURL, imageId, loadCallback) {
        var imageElement = document.createElement('img');

        if (imageURL.lastIndexOf('data:', 0) === 0) {
            //			console.log('About to load image [' + imageId + '] from data URI');
        } else {
            imageElement.crossOrigin = '';
            //			console.log('About to load image [' + imageId + '] from URL: ' + imageURL);
        }

        imageElement.addEventListener('load', function (event) {
            //			console.log('Image [' + imageId + '] loaded');
            loadCallback(imageId, imageElement);
        }, false);

        imageElement.src = imageURL;
    };

    ThreeWrapper.prototype.onWindowClose = function () {
        this.cancelImageSearch();
    };

    ThreeWrapper.prototype.onMouseMoveInternal = function (event) {
        this.mouseMoveOccurred = true;
        var x = event.pageX - this.jQueryCanvas.offset().left;
        var y = event.pageY - this.jQueryCanvas.offset().top;

        this.mousePosition.x = (x / this.canvas.clientWidth) * 2 - 1;
        this.mousePosition.y = -(y / this.canvas.clientHeight) * 2 + 1;

        // do raycast with HUD cube:
        $('#canvas').css('cursor', 'auto');
        this.cubeSideSelected = undefined;
        this.raycaster.setFromCamera(this.mousePosition, this.cameraHud);
        var intersects = this.raycaster.intersectObjects(this.cube.children);
        if (intersects.length > 0) {
            var planeHit = intersects[0].object;
            for (var i = 0; i < this.cubeSides.length; i++) {
                if (this.cubeSides[i] == planeHit) {
                    $('#canvas').css('cursor', 'pointer');
                    this.cubeSideSelected = i;
                    break;
                }
            }
        }

        if (this.cubeSideSelected) {
            this.selectedSpriteId = undefined;
            return;
        }

        // Do raycasting with sprites:
        this.raycaster.setFromCamera(this.mousePosition, this.camera);

        this.selectedSpriteId = undefined;
        var intersects = this.raycaster.intersectObjects(this.spriteContainer.children);
        if (intersects.length > 0) {
            // find first sprite
            var spriteId;
            for (var i = 0; i < intersects.length; i++) {
                var sprite = intersects[i].object;
                spriteId = this.findObjectId(sprite);
                if (spriteId != undefined)
                    break;
            }
            this.selectedSpriteId = spriteId;
        }
    };

    ThreeWrapper.prototype.onMouseMove = function (event) {
        wrapper.onMouseMoveInternal(event);
    };

    ThreeWrapper.prototype.onMouseDownInternal = function () {
        this.cubeSideSelectedOnMouseDown = this.cubeSideSelected;
        this.mouseMoveOccurred = false;
        this.gettingOutOfZoom = false;
        if (this.zoomInObjectId != undefined) {
            this.gettingOutOfZoom = true;
            this.undoZoomIn();
        }
    };

    ThreeWrapper.prototype.onMouseDown = function () {
        wrapper.onMouseDownInternal();
    };

    ThreeWrapper.prototype.onMouseUpInternal = function (event) {
        if (this.cubeSideSelectedOnMouseDown != undefined && this.cubeSideSelected != undefined) {
            this.handleCubeSideSelected(this.cubeSideSelected);
        }

        if (this.gettingOutOfZoom)
            return;
        this.separateAccordingToCamera();

        if (this.mouseMoveOccurred)
            return;
        if (this.selectedSpriteId == undefined)
            return;
        if (this.zoomInObjectId != undefined)
            return;

        var button = event.which || event.button;
        if (button != 1)
            return;

        // zoom-in on object:
        var objectId = this.selectedSpriteId;
        if (!(objectId in this.objectsMap))
            return;

        if ((objectId in this.tweenMap) && objectId == this.zoomInObjectIdPrevious) {
            // we are animating the previous zoom in object back to the original position
            // and we don't want the user to zoom back in on the moving object
            return;
        }

        var sprite = this.objectsMap[objectId];
        this.zoomInObjectId = objectId;
        this.zoomInObjectIdPrevious = objectId;
        this.zoomInObjectPosition = sprite.position.clone();

        var posInFront;

        // TODO the values -23, -30 are device dependent, should be dependent on the size of the window/canvas
        if (sprite.scale.x > sprite.scale.y) {
            posInFront = new THREE.Vector3(0, 0, -1).transformDirection(this.camera.matrixWorld).multiplyScalar(23).add(this.camera.position);
        } else {
            posInFront = new THREE.Vector3(0, 0, -1).transformDirection(this.camera.matrixWorld).multiplyScalar(30).add(this.camera.position);
        }

        if (!this.view3D) {
            // with the orthographic camera we need to scale the sprite as well
            var minWin = Math.min(this.camera2D.right, this.camera2D.top);
            var sideMax = Math.max(sprite.scale.x, sprite.scale.y);
            var factor = 2 * minWin / sideMax;
            this.transitionVector(sprite.scale, new THREE.Vector3(sprite.scale.x * factor, sprite.scale.y * factor, sprite.scale.z), 1000);
        }
        this.transitionForSpacing(sprite, objectId, posInFront, false);
    };

    ThreeWrapper.prototype.onMouseUp = function (event) {
        wrapper.onMouseUpInternal(event);
    };

    ThreeWrapper.prototype.handleCubeSideSelected = function (cubeSide) {
        var camPos4D = new THREE.Vector4(0, 0, this.INITIAL_CAM_POS_FRONT.z);
        switch (cubeSide) {
            case 0:
                var rotWorldMatrix = new THREE.Matrix4();
                rotWorldMatrix.makeRotationAxis(new THREE.Vector3(0, 1, 0), Math.PI / 2);
                camPos4D = camPos4D.applyMatrix4(rotWorldMatrix);
                break;
            case 1:
                var rotWorldMatrix = new THREE.Matrix4();
                rotWorldMatrix.makeRotationAxis(new THREE.Vector3(0, 1, 0), -Math.PI / 2);
                camPos4D = camPos4D.applyMatrix4(rotWorldMatrix);
                break;
            case 2:
                var rotWorldMatrix = new THREE.Matrix4();
                rotWorldMatrix.makeRotationAxis(new THREE.Vector3(1, 0, 0), -Math.PI / 2);
                camPos4D = camPos4D.applyMatrix4(rotWorldMatrix);
                break;
            case 3:
                var rotWorldMatrix = new THREE.Matrix4();
                rotWorldMatrix.makeRotationAxis(new THREE.Vector3(1, 0, 0), Math.PI / 2);
                camPos4D = camPos4D.applyMatrix4(rotWorldMatrix);
                break;
            case 4:
                break;
            case 5:
                var rotWorldMatrix = new THREE.Matrix4();
                rotWorldMatrix.makeRotationAxis(new THREE.Vector3(0, 1, 0), Math.PI);
                camPos4D = camPos4D.applyMatrix4(rotWorldMatrix);
                break;
        }

        this.transitionVector(this.controls.target, new THREE.Vector3(0, 0, 0), 1000);
        this.transitionVector(this.camera.position, new THREE.Vector3(camPos4D.x, camPos4D.y, camPos4D.z), 1000);
    };

    ThreeWrapper.prototype.findObjectId = function (object) {
        var objectId = undefined;
        for (var key in this.objectsMap) {
            if (this.objectsMap.hasOwnProperty(key)) {
                if (this.objectsMap[key] == object) {
                    objectId = key;
                    break;
                }
            }
        }
        return objectId;
    };

    ThreeWrapper.prototype.calculateCurrentPosition = function (spriteId) {
        var sprite = this.objectsMap[spriteId];
        var pos;
        if (this.view3D) {
            pos = this.objectPositionsMap3D[spriteId];
        } else {
            pos = this.objectPositionsMap2D[spriteId];
        }

        var distance = pos.length();
        var factor = this.spacingFactor * this.spacingFactorUser;
        if (this.view3D) {
            if (factor < 1) {
                factor = 1 / (1 + (Math.log(1 / factor)) * (10 / distance));
            } else {
                factor = 1 + (Math.log(factor)) * (10 / distance);
            }
        } else {
            factor = Math.round(Math.log(factor) / Math.log(8));
            factor = Math.exp(factor / 10);
        }

        return pos.clone().multiplyScalar(factor);
    };

    ThreeWrapper.prototype.calculateCurrentPositionWithCamera = function (key) {
        var shouldBePosition = this.calculateCurrentPosition(key);

        var sprite = this.objectsMap[key];
        var camPos = this.camera.position;

        if (this.view3D) {
            // 3D mode
            var currentPos = sprite.position;

            var direction = new THREE.Vector3().subVectors(currentPos, camPos);
            var direction2 = new THREE.Vector3().subVectors(shouldBePosition, camPos);

            var distance = direction.length();
            var distance2 = direction2.length();

            if (distance < this.MINIMUM_CAMERA_DISTANCE || distance2 < this.MINIMUM_CAMERA_DISTANCE) {
                if (distance2 < this.MINIMUM_CAMERA_DISTANCE + this.MINIMUM_CAMERA_DISTANCE / 4) {
                    // move away object along camera -> object vector
                    var moveAwayDistance = this.MINIMUM_CAMERA_DISTANCE + this.MINIMUM_CAMERA_DISTANCE * 0.05;
                    return direction.normalize().multiplyScalar(moveAwayDistance).add(camPos);
                }
            }

            /* else -> we restore the shouldBePosition */
            return shouldBePosition;
        } else {
            // 2D mode
            var camDistance = camPos.z;
            var restore = camDistance > 200;
            var distance = shouldBePosition.length();

            if (!restore) {
                var scaling = Math.min(700, Math.max(distance, Math.pow(distance * 100 / camDistance, 1.2)));
                return shouldBePosition.normalize().multiplyScalar(scaling);
            }
            return shouldBePosition;
        }
    };

    ThreeWrapper.prototype.separateAccordingToCamera = function () {
        var camPos = this.camera.position.clone();
        var target = this.controls.target;

        for (var key in this.objectsMap) {
            if (this.objectsMap.hasOwnProperty(key)) {
                var sprite = this.objectsMap[key];
                var position = this.calculateCurrentPositionWithCamera(key);
                this.transitionForSpacing(sprite, key, position, true);
            }
        }
    };

    ThreeWrapper.prototype.zooming = function (delta, speed) {
        this.undoZoomIn();

        var camPos = this.camera.position;
        var target = this.controls.target;
        var direction = new THREE.Vector3().subVectors(target, camPos);
        direction.normalize();
        var normDir = direction.clone();
        var scaling = 25 * speed;
        direction.multiplyScalar(scaling);

        if (delta < 0) {
            direction.multiplyScalar(-1);
        }

        var minObjectDistance = Number.MAX_VALUE;

        for (var key in this.objectsMap) {
            if (this.objectsMap.hasOwnProperty(key)) {
                var sprite = this.objectsMap[key];
                var pos = this.calculateCurrentPosition(key);
                minObjectDistance = Math.min(minObjectDistance, camPos.distanceTo(pos));
            }
        }

        if (this.view3D) {
            var oldTarget = this.controls.target;
            var targetDistance = new THREE.Vector3().subVectors(camPos, oldTarget).length();
            var isMode2 = targetDistance < 0.5;

            var camToTarget = new THREE.Vector3().subVectors(oldTarget, camPos);
            var cosAngle = camPos.dot(camToTarget);

            this.camera.position.set(camPos.x + direction.x, camPos.y + direction.y, camPos.z + direction.z);
            camPos = this.camera.position.clone();

            var switchToMode2 = minObjectDistance < this.MINIMUM_CAMERA_DISTANCE / 1.5 || targetDistance < 50;
            var switchToMode1 = delta < 0 && minObjectDistance > this.MINIMUM_CAMERA_DISTANCE && cosAngle < -0.5;

            //console.log('is mode2: ' + isMode2 + ', switch to 1: ' + switchToMode1 + ', switch to 2: ' + switchToMode2);
            if (switchToMode1 && isMode2) {
                // if we are zooming out and are far away from the origin and not looking at (0,0,0) then we recenter the camera to (0,0,0)
                var newTarget = new THREE.Vector3(0, 0, 0);
                this.transitionCamera(this.controls.target, newTarget, 1000);
            } else if (switchToMode2 || isMode2) {
                var targetDistance = 0.2;

                // when we get near, then we set the target near the camera to allow rotating around the camera
                var newTarget = new THREE.Vector3(camPos.x + normDir.x * targetDistance, camPos.y + normDir.y * targetDistance, camPos.z + normDir.z * targetDistance);
                this.controls.target.set(newTarget.x, newTarget.y, newTarget.z);
            }
        } else {
            var newCamPos = new THREE.Vector3().addVectors(camPos, direction);
            if (delta >= 0 && newCamPos.z < 50)
                return;

            this.camera.position.set(camPos.x + direction.x, camPos.y + direction.y, camPos.z + direction.z);

            var zoomFactor = 1 + (0.1 * speed);
            if (delta < 0) {
                zoomFactor = 1 / zoomFactor;
            }

            // keep the backround grid at a fixed size:
            this.backgroundGrid.scale.set(this.backgroundGrid.scale.x / zoomFactor, this.backgroundGrid.scale.y / zoomFactor, 1);

            this.camera2D.left /= zoomFactor;
            this.camera2D.right /= zoomFactor;
            this.camera2D.top /= zoomFactor;
            this.camera2D.bottom /= zoomFactor;

            //(<any> this.camera2D).zoom *= zoomFactor;
            this.camera2D.updateProjectionMatrix();
        }
        this.separateAccordingToCamera();
    };

    ThreeWrapper.prototype.onMouseWheelInternal = function (event) {
        // HACK OrbitControls, such that the target moves along with the mouse wheel:
        var delta = 0;
        if (event.wheelDelta !== undefined) {
            delta = event.wheelDelta;
        } else if (event.detail !== undefined) {
            delta = -event.detail;
        }

        this.zooming(delta, 1);
    };

    ThreeWrapper.prototype.onMouseWheel = function (event) {
        wrapper.onMouseWheelInternal(event);
    };

    ThreeWrapper.prototype.onTouchStart = function (event) {
        if (event.touches.length != 1) {
            wrapper.undoZoomIn();
        }

        if (event.touches.length == 2) {
            var dx = event.touches[0].pageX - event.touches[1].pageX;
            var dy = event.touches[0].pageY - event.touches[1].pageY;
            var distance = Math.sqrt(dx * dx + dy * dy);
            wrapper.dollyStart.set(0, distance);
        }
    };

    ThreeWrapper.prototype.onTouchMove = function (event) {
        if (event.touches.length == 2) {
            var dx = event.touches[0].pageX - event.touches[1].pageX;
            var dy = event.touches[0].pageY - event.touches[1].pageY;
            var distance = Math.sqrt(dx * dx + dy * dy);

            wrapper.dollyEnd.set(0, distance);
            wrapper.dollyDelta.subVectors(wrapper.dollyEnd, wrapper.dollyStart);

            wrapper.zooming(wrapper.dollyDelta.y, wrapper.ZOOM_SPEED_ON_MOBILE);

            wrapper.dollyStart.copy(wrapper.dollyEnd);
        }
    };

    ThreeWrapper.prototype.onAnythingToUndoZoomIn = function () {
        wrapper.undoZoomIn();
    };

    ThreeWrapper.prototype.undoZoomIn = function () {
        if (this.zoomInObjectId == undefined)
            return;
        var sprite = this.objectsMap[this.zoomInObjectId];
        var restoredPos = this.zoomInObjectPosition;

        // restore sprite scale:
        var previousScale = this.objectScaleMap[this.zoomInObjectId];
        this.transitionVector(sprite.scale, previousScale, 1000);

        this.transitionForSpacing(sprite, this.zoomInObjectId, restoredPos, false);
        this.zoomInObjectId = undefined;
    };

    ThreeWrapper.prototype.updateSelectionBorder = function () {
        if (!this.showShadowsForThisSearch)
            return;

        var spriteId = this.selectedSpriteId;
        if (spriteId == this.selectedSpriteIdPrev) {
            if (spriteId != undefined) {
                // update position/scale/rotation
                var sprite = this.objectsBackMap[spriteId];
                this.selectionBorder.position.copy(sprite.position);
                this.selectionBorder.quaternion.copy(sprite.quaternion);
                this.selectionBorder.scale.copy(sprite.scale);

                if (this.zoomInObjectId != undefined) {
                    // restore old state
                    this.selectionBorder.visible = false;
                    var previous = this.objectsBackMap[this.selectedSpriteIdPrev];
                    previous.material.uniforms.shadowTex = { type: 't', value: this.shadowTexture };
                    previous.material.needsUpdate = true;
                }
            }
            return;
        }

        if (this.selectedSpriteIdPrev != undefined) {
            // restore old state
            var previous = this.objectsBackMap[this.selectedSpriteIdPrev];
            previous.material.uniforms.shadowTex = { type: 't', value: this.shadowTexture };
            previous.material.needsUpdate = true;
        }
        this.selectedSpriteIdPrev = spriteId;

        if (spriteId == undefined || this.zoomInObjectId != undefined) {
            this.selectionBorder.visible = false;
            return;
        }

        if (!(spriteId in this.objectsMap))
            return;

        var backsprite = this.objectsBackMap[spriteId];
        backsprite.material.uniforms.shadowTex = { type: 't', value: this.selectionTexture };
        backsprite.material.needsUpdate = true;

        this.scene.remove(this.selectionBorder);
        this.selectionBorder = backsprite.clone(undefined, false);
        this.selectionBorder.visible = true;
        this.scene.add(this.selectionBorder);
    };

    ThreeWrapper.prototype.transitionCamera = function (fromPosition, toPosition, duration) {
        if (fromPosition.x == toPosition.x && fromPosition.y == toPosition.y && fromPosition.z == toPosition.z)
            return;

        new TWEEN.Tween(fromPosition).to({
            x: toPosition.x,
            y: toPosition.y,
            z: toPosition.z
        }, duration).easing(TWEEN.Easing.Circular.InOut).start();
    };

    ThreeWrapper.prototype.transitionVector = function (fromPosition, toPosition, duration) {
        if (!this.showAnimations) {
            fromPosition.copy(toPosition);
            return;
        }
        var diffInPos = new THREE.Vector3().subVectors(fromPosition, toPosition).length();
        if (diffInPos < 0.001)
            return;

        new TWEEN.Tween(fromPosition).to({
            x: toPosition.x,
            y: toPosition.y,
            z: toPosition.z
        }, duration).easing(TWEEN.Easing.Circular.InOut).start();
    };

    ThreeWrapper.prototype.transitionForSpacing = function (object, objectId, toPosition, excludeZoomObject) {
        this.transitionDuration(object, objectId, toPosition, 1000, excludeZoomObject);
    };

    ThreeWrapper.prototype.transitionDuration = function (object, objectId, toPosition, duration, excludeZoomObject) {
        var fromPosition = object.position;

        if (excludeZoomObject && objectId == this.zoomInObjectId)
            return;

        // check if we need to create a transition:
        var diffInPos = new THREE.Vector3().subVectors(fromPosition, toPosition).length();
        if (diffInPos < 0.001)
            return;

        if (!this.showAnimations) {
            fromPosition.copy(toPosition);
            return;
        }

        if (this.tweenMap.hasOwnProperty(objectId)) {
            // stop current tween and start a new tween from the current position to the end position,
            // but we don't want to use exponential movement in the initial phase (only at the end)
            // we are assuming we are already moving...
            var tween = this.tweenMap[objectId];
            var currentPos = object.position.clone();

            tween.stop();
            var tween2 = new TWEEN.Tween(fromPosition).to({
                x: toPosition.x,
                y: toPosition.y,
                z: toPosition.z
            }, duration);
            tween2.easing(TWEEN.Easing.Exponential.Out);
            tween2.onComplete(function (thetween) {
                delete wrapper.tweenMap[objectId]; // remove from map
            });

            tween2._valuesStart = tween._valuesEnd;

            this.tweenMap[objectId] = tween2;
            tween2.start();
        } else {
            var tween = new TWEEN.Tween(fromPosition).to({
                x: toPosition.x,
                y: toPosition.y,
                z: toPosition.z
            }, duration);
            tween.easing(TWEEN.Easing.Exponential.InOut);
            tween.onComplete(function (thetween) {
                delete wrapper.tweenMap[objectId]; // remove from map
            });

            this.tweenMap[objectId] = tween;
            tween.start();
        }
    };

    ThreeWrapper.prototype.transitionRandom = function (object, objectId, toPosition) {
        var duration = 2000;
        this.transitionDuration(object, objectId, toPosition, Math.random() * duration + duration, true);
    };

    ThreeWrapper.prototype.updateForResize = function () {
        var canvas = this.canvas;
        var container = this.canvasContainer;
        this.jQueryContainer.css('top', this.topControls.clientHeight); // set the 'top' property of css to the current height of the controls

        var width = container.clientWidth;
        var height = container.clientHeight;
        if (this.lastContainerWidth == width && this.lastContainerHeight == height)
            return;

        // update WebGL canvas and properties:
        this.lastContainerWidth = width;
        this.lastContainerHeight = height;

        this.renderer.setSize(width, height); // MUST USE this method instead of setting width/height on canvas directly!
        var aspect = width / height;

        this.renderer.setViewport(0, 0, width, height);

        this.camera3D.aspect = aspect;
        this.camera3D.updateProjectionMatrix();

        this.camera2D.top = this.camera2D.right / aspect;
        this.camera2D.bottom = this.camera2D.left / aspect;
        this.camera2D.updateProjectionMatrix();

        this.cameraHud.aspect = aspect;
        this.cameraHud.updateProjectionMatrix();
        this.cameraHud.setViewOffset(width / 4, height / 4, 0, 0, width, height);
    };

    ThreeWrapper.prototype.updateScene = function () {
        this.updateForResize();
        TWEEN.update();

        // all sprites (= planes) should face the camera:
        var camPos = this.camera.position.clone();

        // control camera position of HUD:
        var camTarget = this.controls.target.clone();
        var direction = camPos.clone().sub(camTarget).normalize();

        var camHudPos = camPos.clone().normalize();
        this.cameraHud.position.copy(direction.multiplyScalar(this.HUD_CUBE_SIZE));
        this.cameraHud.lookAt(new THREE.Vector3(0, 0, 0));

        for (var key in this.objectsMap) {
            if (this.objectsMap.hasOwnProperty(key)) {
                var sprite = this.objectsMap[key];
                sprite.quaternion.copy(this.camera.quaternion);

                // sprites with shadow need to be at the same location (and suitable scaling for ortho-camera) as the sprites:
                var spriteBack = this.objectsBackMap[key];
                spriteBack.scale.copy(sprite.scale);
                if (this.showShadowsForThisSearch) {
                    var additionalScale = this.objectsBackScaleMap[key];
                    spriteBack.scale.multiply(additionalScale);
                }
                spriteBack.position.copy(sprite.position);
                spriteBack.quaternion.copy(this.camera.quaternion);
            }
        }

        this.updateSelectionBorder();

        this.controls.update();

        var context = this.renderer.getContext();

        this.spriteBackContainer.visible = false;

        var selectionBorderVisible = this.selectionBorder.visible;

        // 1. render only the sprites to the stencil buffer
        this.spriteContainer.visible = true;
        this.spriteBackContainer.visible = false;
        this.axesContainer.visible = false;
        this.groundPlane.visible = false;
        this.selectionBorder.visible = false;
        this.backgroundGrid.visible = false;

        context.colorMask(false, false, false, false);
        context.depthMask(false);
        context.enable(context.STENCIL_TEST);
        context.stencilFunc(context.ALWAYS, 1, 0xff); // always passes and write 1s
        context.stencilOp(context.KEEP, context.KEEP, context.REPLACE);
        context.stencilMask(0xff);

        this.renderer.clear(false, true, true);
        this.renderer.render(this.scene, this.camera);

        // 2. now render the sprites with shadow, but only allow writing where the stencil buffer is 1
        // (the order is sorted automatically by three.js and the depths are given by the sprites we are rendering now)
        this.spriteBackContainer.visible = true;
        this.spriteContainer.visible = false;

        context.colorMask(true, true, true, true);
        context.clearColor(1, 1, 1, 0);
        context.depthMask(true);
        this.renderer.clear(true, false, false); // clear with clearcolor
        context.stencilFunc(context.EQUAL, 1, 0xff); // only where the stencil has been set to 1
        context.stencilMask(0x00); // to avoid modification to the stencil buffer

        this.renderer.render(this.scene, this.camera);

        // 3. now render the rest (i.e. the axes and the plane)
        this.spriteBackContainer.visible = false;
        this.groundPlane.visible = this.view3D;
        this.axesContainer.visible = true;
        this.selectionBorder.visible = selectionBorderVisible;
        this.backgroundGrid.visible = !this.view3D;

        context.disable(context.STENCIL_TEST);
        this.renderer.render(this.scene, this.camera);

        // FINAL: render HUD
        this.renderer.clear(false, true, false); // clear depth - to keep HUD always on top
        this.renderer.render(this.sceneHud, this.cameraHud);
    };
    return ThreeWrapper;
})();

var render = function () {
    requestAnimationFrame(render);
    wrapper.updateScene();
};

var doImageSearch = function () {
    var query = $('#txtSearch').val();
    var count = $('#numImgCount').val();
    console.log('Searching for "' + query + '"');
    wrapper.triggerImageSearch(query, count);
};

var doCancelSearch = function () {
    wrapper.cancelImageSearch();
};

var doClearResults = function () {
    wrapper.reset();
};

var doIncreaseSpacing = function () {
    wrapper.increaseSpacing();
};

var doDecreaseSpacing = function () {
    wrapper.decreaseSpacing();
};

var doSwitch2d3d = function () {
    wrapper.switch2d3d();
};

var doChangeAnimations = function () {
    var showAnimations = document.getElementById('checkAnim').checked;
    wrapper.showAnimations = showAnimations;
};

var doChangeShadows = function () {
    var showShadows = document.getElementById('checkShadow').checked;
    wrapper.showShadows = showShadows;
    $("#shadowHint").css('visibility', 'visible');
};

window.onload = function () {
    var el = document.getElementById('content');
    wrapper = new ThreeWrapper(el);
    wrapper.init();
    render();
};
//# sourceMappingURL=app.js.map
