﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using ICSV.Interface.Types;

namespace ICSV.Interface.Operations
{
	/// <summary>
	/// DTO for getting image results for the given search id.
	/// The result will be of <see cref="GetImageResultsResponse"/>
	/// </summary>
	public class GetImageResults
	{
		/// <summary>
		/// The search id for the results
		/// </summary>
		public int ForSearchId { get; set; }
	}

	/// <summary>
	/// The response to an image results request
	/// </summary>
	public class GetImageResultsResponse
	{
		/// <summary>
		/// The search id for the results
		/// </summary>
		public int SearchId { get; set; }

		/// <summary>
		/// The resulting list of images
		/// </summary>
		public ArrayOfImageData Images { get; set; }

		/// <summary>
		/// Whether or not it is the final result.
		/// True, if there are no more image results available.
		/// False, if the server can deliver more than that in subsequent result requests.
		/// </summary>
		public bool IsFinalResult { get; set; }

		/// <summary>
		/// The list of axes for the images for the 2D representation.
		/// </summary>
		public ArrayOfAxisUpdate Axes2D { get; set; }

		/// <summary>
		/// The list of axes for the images for the 3D representation.
		/// </summary>
		public ArrayOfAxisUpdate Axes3D { get; set; }
	}
}
