﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace ICSV.Interface.Operations
{
	/// <summary>
	/// The request for starting a new search of images.
	/// The response will be <see cref="SearchImagesResponse"/>, which will deliver a search id for further request.
	/// </summary>
	public class SearchImages
	{
		/// <summary>
		/// The query string indicates what to search for.
		/// </summary>
		/// <remarks>
		/// A comma-delimited list of tags. Photos with one or more of the tags listed will be returned.
		/// You can exclude results that match a term by prepending it with a - character.
		/// </remarks>
		public string ForQuery { get; set; }

		/// <summary>
		/// The maximum number of images the server should look for.
		/// </summary>
		public int PhotoCount { get; set; }
	}

	/// <summary>
	/// The response of a new image search request.
	/// </summary>
	public class SearchImagesResponse
	{
		/// <summary>
		/// The search id, used to reference subsequent inquiries.
		/// <seealso cref="GetImageResults"/>
		/// </summary>
		public int SearchId { get; set; }
	}
}
