﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace ICSV.Interface.Types
{
	/// <summary>
	/// An update for an axis. Included in the list of an <see cref="GetImageResultsResponse"/> object.
	/// </summary>
	/// <remarks>
	/// An axis consists of two locations in space (either 2D or 3D) and a color for each point.
	/// The axis is considered to be connecting these two points.
	/// </remarks>
	public class AxisUpdate
	{
		/// <summary>
		/// X coordinate of point 1.
		/// </summary>
		public double XCoordP1 { get; set; }
		
		/// <summary>
		/// Y coordinate of point 1.
		/// </summary>
		public double YCoordP1 { get; set; }
		
		/// <summary>
		/// Z coordinate of point 1. Is zero for 2D representation.
		/// </summary>
		public double ZCoordP1 { get; set; }

		/// <summary>
		/// X coordinate of point 2.
		/// </summary>
		public double XCoordP2 { get; set; }
		
		/// <summary>
		/// Y coordinate of point 2.
		/// </summary>
		public double YCoordP2 { get; set; }
		
		/// <summary>
		/// Z coordinate of point 2. Is zero for 2D representation.
		/// </summary>
		public double ZCoordP2 { get; set; }

		/// <summary>
		/// The color of point 1 in the format: (R shifted by 16) + (G shifted by 8) + B
		/// </summary>
		public int ColorP1 { get; set; }

		/// <summary>
		/// The color of point 2 in the format: (R shifted by 16) + (G shifted by 8) + B
		/// </summary>
		public int ColorP2 { get; set; }		
	}
}
